/*
 2A0x.c
  N2A03/N2A07 CPU emulation; #include's ops6510.h

  The N2A03 CPU is a modified NMOS 6502, adding 6510 instructions,
   removing decimal code (ADC/SBC are always computed in binary mode).
  Besides the rate divider, the CPUs for the 2A03 and 2A07 are identical.

 * MacFCEU 0.9d6
 *  For testing: nestest.nes, NEStress.nes, cpu.nes / blargg's CPU tests.
 *
 * Copyright notice for this file:
 *  x6502.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "types.h"
#include "x6502.h"
#include "FCEU.h"
#include "debug.h"
#include "fsound.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

/* ---- globals */

X6502U		X6502F;
uint32		timestamp;

uint8		ZNTable[256];
extern int32	XFStatL;

uint8		CycTable[256] =
{ /* longest time: 8 cycles (line %xxx00011) */
/*0x00*/ 7,6,2,8,3,3,5,5,3,2,2,2,4,4,6,6,
/*0x10*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x20*/ 6,6,2,8,3,3,5,5,4,2,2,2,4,4,6,6,
/*0x30*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x40*/ 6,6,2,8,3,3,5,5,3,2,2,2,3,4,6,6,
/*0x50*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x60*/ 6,6,2,8,3,3,5,5,4,2,2,2,5,4,6,6,
/*0x70*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0x80*/ 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
/*0x90*/ 2,6,2,6,4,4,4,4,2,5,2,5,5,5,5,5,
/*0xA0*/ 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
/*0xB0*/ 2,5,2,5,4,4,4,4,2,4,2,4,4,4,4,4,
/*0xC0*/ 2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
/*0xD0*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
/*0xE0*/ 2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6, /*$E2 was 3 (0.98.13) */
/*0xF0*/ 2,5,2,8,4,4,6,6,2,4,2,7,4,4,7,7,
};

/* ---- code */

void FP_FASTAPASS(1) (*MapIRQHook)(int a);
void FP_FASTAPASS(1) (*DebugStepHook)(int c);

/* use these for reads, optimizing out zero extension */
#define RPC		X6502F.u32.PC
#define RA		X6502F.u32.A
#define RX		X6502F.u32.X

#define _A		XFrame.A
#define _X		XFrame.X
#define _Y		XFrame.Y
#define _S		XFrame.S
#define _P		XFrame.P
#define _PI		XFrame.BP
#define _count		XFrame.count
#define _tcount		XFrame.tcount
#define _intlow		XFrame.intlow

#define _ADDCYC(x)	(_count-=x*CPU_SUBCYC,_tcount+=x,timestamp+=x)
/* minus interrupt edge */
#define _ADDCYCI(x)	(_count-=x*CPU_SUBCYC-CPU_NMIEDGE,_tcount+=x,timestamp+=x)
#define ADDCYC		_ADDCYC	/*C language takes care of the math*/

/* I've managed to get all occurences of the read/write macros to use a variable, no expressions */
#define RMem(A)		ARead[A](A)	/*_DB=*/
#define XDMR(A)		(_ADDCYC(1),RMem(A))
#define _RMem16(A)	(XFrame.IAx3=ARead[A](A),XFrame.IAx2=ARead[(A+1)&0xFFFF]((A+1)&0xFFFF), \
	X6502F.u32.IA)
#define WMem(A,V)	BWrite[A](A,V)

uint8 RMemCheck(unsigned A);
void WMemCheck(unsigned A,
	uint8 V);

static uint8 INLINE RMemCheck(unsigned A)
{
	if (A>0xFFFF)
	{
		FCEU_PrintError("Bug Alert!  RMemCheck got a bad address (%d).",
			A);
		return A;
	}
	return RMem(A);
}

static void INLINE WMemCheck(unsigned A,
	uint8 V)
{
	if (A>0xFFFF)
	{
		FCEU_PrintError("Bug Alert!  WMemCheck got a bad address (%d).",
			A);
		return;
	}
	WMem(A,V);
}

uint8 FASTAPASS(1) X6502_DMR(uint32 A)
{
	if (A>0xFFFF)
	{
		FCEU_PrintError("Bug Alert!  X6502_DMR got a bad address (%d).",
			A);
		return A;
	}
	return XDMR(A);
}

void FASTAPASS(2) X6502_DMW(uint32 A,
	uint8 V)
{
	if (A>0xFFFF)
	{
		FCEU_PrintError("Bug Alert!  X6502_DMW got a bad address (%d).",
			A);
		return;
	}
	_ADDCYC(1);
	WMem(A,V);
}

DECLFW(DMA4014);
DECLFW(DMA4014) /* copy page to 2004h */
{
	int		c;

	if ((int64lo(timestampbase)+timestamp)&1)
		_ADDCYC(1);
	c = 256, A = (uint32)V<<8;
	do
	{
		V = XDMR(A);
		_ADDCYC(1);
		WMem(0x2004,V);
	} while(++A,--c);
}

void FASTAPASS(1) X6502_IRQBegin(int w)
{
	XFrame.intlow|= w;
}

void FASTAPASS(1) X6502_IRQEnd(int w)
{
	XFrame.intlow&=~w;
}

/* added; 2011.02.13, 2015.09.20 */
/* the B flag is used in passing here (so it gets erased next instr.) to delay NMIs in PPU.c */
#define IRQVector()	_RMem16(0xFFFE)
#define NMIVector()	_RMem16(0xFFFA)
#define IRSRVector()	(*(int16*)&_PI|=B_FLAG*0x101, /* current set for non-BRK transfer */ \
	(XFrame.NMISR|XFrame.intlow<<7-kIQNMI2)&0x80 ? NMIVector():IRQVector())

/* full mode can now do stack effects; 2010.04.11 */
/* renamed from PUSH/POP to avoid matching name conflicts; 2010.04.21 */
#define PSH(V)		do{WRAM(0x100+_S,V); --_S;}while(0)

/* parameters only used once, only _P and ZNTable are used during */
#define X_ZNT(v)	_P|=ZNT[v]
#define X_ZN(v)		_P&=~(N_FLAG|Z_FLAG),X_ZNT(v)

/* extra cycle for page update (8 bit carry) microcode */

#define JR(cond)	if(!(cond)) {++_PC,_PC&=0xFFFF; break;} else goto jr_destin;
#define JR_DESTIN	jr_destin: {x=(int8)XDMR(_PC), \
	t=++_PC,_PC+=x,_PC&=0xFFFF,t^=_PC; \
	if(t&=0x100) { \
		_ADDCYC(1); /* this is not a dead cycle! */ \
		if(!(_PI&I_FLAG) && !(XFStatL&=0x80FFFFFEu)) \
			FrameSoundIRQ(); \
	} \
	break;}

#define LDA		X_ZN(RA=x)
#define LDX		X_ZN(RX=x)
#define LDY		X_ZN(_Y=x)
#define LAX		X_ZN(RA=RX=x) /* differing implementations (not for use on C64-II) */
#define LAS		(x&=_S,_S=x,LAX)

/* bit-wise operations */
#define SAX		(RA&RX)

#define ANA		X_ZN(RA&=x)
#define BIT		_P&=~(N_FLAG|V_FLAG|Z_FLAG),_P|=(!(x&RA))*Z_FLAG,_P|=x&(N_FLAG|V_FLAG)
#define EOA		X_ZN(RA^=x)
#define ORA		X_ZN(RA|=x)

/* arithmetic operations -- note that N2A0x's have no decimal mode functioning microcode */
#define ADC		(t=_P&C_FLAG,_P&=~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG),t+=RA,t+=x, \
	_P|=t>>8&C_FLAG,x^=RA,_P|=((t^RA)&~x)>>1&V_FLAG,x^=RA,X_ZNT(RA=t&=0xFF))
#define SBC		(x^=0xFF,ADC) //modifies x (don't use in SBC-extensions, like ISB)

#define CMPL(a,b)	(t=a,t-=b,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=~(t>>8)&C_FLAG,X_ZNT(t&=0xFF))

#define ANC		(x&=RA,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x>>7,X_ZNT(RA=x))

/* Special undocumented operation.  Very similar to CMP. */
#define AXS		(t=RA&RX,t-=x, \
	_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=~(t>>8)&C_FLAG,X_ZNT(RX=t&=0xFF))

#define CPA		CMPL(RA,x)
#define CPX		CMPL(RX,x)
#define CPY		CMPL(_Y,x)

/* incremental ops */
#define DEC		(--x,X_ZN(x&=0xFF))
#define INC		(++x,X_ZN(x&=0xFF))
#define DCP		(--x,x&=0xFF,CPA)
#define ISB		(x=(uint8)~(x+1),ADC,x^=0xFF)

#define ASL		(x+=x,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x>>8,X_ZNT(x&=0xFF))
#define LSR		(_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x&C_FLAG,X_ZNT(x>>=1))
#define ROL		(x+=x,x+=_P&C_FLAG,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x>>8,X_ZNT(x&=0xFF))
#define ROR		(x+=(_P&C_FLAG)<<8,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x&C_FLAG,X_ZNT(x>>=1))

/* arithmetic and-rotate-right - corrected */
#define ARR		(x&=RA,t=x,x=x>>1|(uint8)(_P<<7),_P&=~(N_FLAG|C_FLAG|Z_FLAG|V_FLAG), \
	_P|=x>>6&C_FLAG,_P|=(t^x)&V_FLAG,X_ZNT(RA=x))
/* and-shift-right - fixed */
#define ASR		(x&=RA,_P&=~(N_FLAG|C_FLAG|Z_FLAG),_P|=x&C_FLAG,X_ZNT(RA=x>>1))

#define RLA		(x+=x,x+=_P&C_FLAG,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x>>8,X_ZNT(RA&=x))
#define RRA		(x+=(_P&C_FLAG)<<8,_P&=~C_FLAG,_P|=x&C_FLAG,x>>=1,ADC)
#define SLO		(x+=x,_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x>>8,x&=0xFF,ORA)
#define SRE		(_P&=~(N_FLAG|Z_FLAG|C_FLAG),_P|=x&C_FLAG,X_ZNT(RA^=x>>=1))

/* "address high" ops */
/* a plus one for precalculated boundary crossing */
#define SHA		(t=_Y+(uint8)A,A&=0xFF00,RMem(A|=(uint8)t), x=RA&RX&((A>>8)+1), \
	t>255 ? (A=(uint8)A|x<<8):0,x)
#define SHY		(t=RX+(uint8)A,A&=0xFF00,RMem(A|=(uint8)t), x=_Y&((A>>8)+1), \
	t>255 ? (A=(uint8)A|x<<8):0,x)
#define SHX		(t=_Y+(uint8)A,A&=0xFF00,RMem(A|=(uint8)t), x=RX&((A>>8)+1), \
	t>255 ? (A=(uint8)A|x<<8):0,x)
#define SHS		(t=_Y+(uint8)A,A&=0xFF00,RMem(A|=(uint8)t), x=(_S=_A&_X)&((A>>8)+1), \
	t>255 ? (A=(uint8)A|x<<8):0,x)

/* om6502 version of ANE */
#define ANE		(x&=RX,ANA)

/* interrupts */

/*
 According to m6502/om6502 in MAME, in the event an interrupt fires during
a BRK, or any instruction, the B flag is cleared and the PC is not incremented.
During prefetch, an interrupt will also replace the op with $00 (implicit BRK).
 BRK calls the interrupt/ISR microcode, so the effect on an NMI with the
instruction is the same after IRQ is triggered - delayed.
*/
#define BRK		{++_PC; PSH(_PC>>8); PSH(_PC); PSH(_P|M_FLAG|B_FLAG); \
	*(int16*)&_PI|=(0x101*I_FLAG),_PC=IRSRVector();}

/*
 According to Visual6502, a B flag is just an indicator in the ISR write,
depending on IRQ/BRK, and not maintained as part of the P register.
*/
#define RTI		(void)(_PI=_P=PUL(),_PC=PUL(),_PC|=PUL()<<8)

/* JMP indirect absolute (the only use of this special mode) */
#define JIN		{GetAB(A); _PC=RMem(A); \
	if(!(++A&0xFF)) \
		A-=0x100; /* 8 bit wrap */ \
	_PC|=RMem(A)<<8; \
}

/* GetABIRD and GetIYRD up to FCEUX 2.1.2 does t^=0x100, not t-=0x100. */

/* Absolute */
#define GetAB(d)	(d=RMem(_PC),++_PC,_PC&=0xFFFF,d+=RMem(_PC)<<8,++_PC,_PC&=0xFFFF)

/* Absolute Indexed (for reads) */
#define GetABIRD(d,i)	{t=RMem(_PC)+i,++_PC,_PC&=0xFFFF,d=t+(RMem(_PC)<<8),++_PC,_PC&=0xFFFF; \
	if(t&=0x100) t=d-0x100,d&=0xFFFF,XDMR(t);}

/* Absolute Indexed (for writes and rmws) */
#define GetABIWR(d,i)	(t=RMem(_PC)+i,++_PC,_PC&=0xFFFF,d=t+(RMem(_PC)<<8),++_PC,_PC&=0xFFFF, \
	t&=0x100,t=d-t,d&=0xFFFF,RMem(t))

#define GetIX(d)	(t=RMem(_PC),t+=RX,t&=0xFF,++_PC,_PC&=0xFFFF, \
	d=RRAM(t),++t,t&=0xFF,d+=RRAM(t)<<8)

/* Indirect Indexed (for reads) */
#define GetIYRD(d)	{d=RMem(_PC),++_PC,_PC&=0xFFFF,t=RRAM(d)+_Y,++d,d&=0xFF,d=t+(RRAM(d)<<8); \
	if(t&=0x100) t=d-0x100,d&=0xFFFF,XDMR(t);}

/* Indirect Indexed (for writes and rmws) */
#define GetIYWR(d)	(d=RMem(_PC),++_PC,_PC&=0xFFFF,t=RRAM(d)+_Y,++d,d&=0xFF,d=t+(RRAM(d)<<8), \
	t&=0x100,t=d-t,d&=0xFFFF,RMem(t))

/* indirect for SHA - Y not added here */
#define GetIHWR(d)	(d=RMem(_PC),++_PC,_PC&=0xFFFF,t=RRAM(d),++d,d&=0xFF,d=t+(RRAM(d)<<8))

/* Now come the macros to wrap up all of the above stuff addressing mode functions
   and operation macros.  Note that operation macros will always operate(redundant
   redundant) on the variable "x".
*/
#define RMW_A(op)	{IMP; x=RA; op; _A=x; break;} /* Meh... */
#define RMW_AB(op)	{GetAB(A),x=RMem(A); WMem(A,x); op; WMem(A,x); break;}
#define RMW_ABX(op)	{GetABIWR(A,RX); x=RMem(A); WMem(A,x); op; WMem(A,x); break;}
#define RMW_ABY(op)	{GetABIWR(A,_Y); x=RMem(A); WMem(A,x); op; WMem(A,x); break;}
#define RMW_IX(op)	{GetIX(A),x=RMem(A); WMem(A,x); op; WMem(A,x); break;}
#define RMW_IY(op)	{GetIYWR(A),x=RMem(A); WMem(A,x); op; WMem(A,x); break;}
#define RMW_ZP(op)	{A=RMem(_PC),++_PC,_PC&=0xFFFF,x=RRAM(A); op; WRAM(A,x); break;}
#define RMW_ZPX(op)	{A=(uint8)(_X+RMem(_PC)),++_PC,_PC&=0xFFFF,x=RRAM(A); op; WRAM(A,x); break;}

#define LD_IM(op)	{x=RMem(_PC),++_PC,_PC&=0xFFFF; op; break;}
#define LD_ZP(op)	{A=RMem(_PC),++_PC,_PC&=0xFFFF,x=RRAM(A); op; break;}
#define LD_ZPX(op)	{A=(uint8)(_X+RMem(_PC)),++_PC,_PC&=0xFFFF,x=RRAM(A); op; break;}
#define LD_ZPY(op)	{A=(uint8)(_Y+RMem(_PC)),++_PC,_PC&=0xFFFF,x=RRAM(A); op; break;}
#define LD_AB(op)	{GetAB(A),x=RMem(A); op; break;}
#define LD_ABX(op)	{GetABIRD(A,RX); x=RMem(A); op; break;}
#define LD_ABY(op)	{GetABIRD(A,_Y); x=RMem(A); op; break;}
#define LD_IX(op)	{GetIX(A),x=RMem(A); op; break;}
#define LD_IY(op)	{GetIYRD(A); x=RMem(A); op; break;}

#define ST_ZP(r)	{A=RMem(_PC),++_PC,_PC&=0xFFFF; WRAM(A,r); break;}
#define ST_ZPX(r)	{A=(uint8)(_X+RMem(_PC)),++_PC,_PC&=0xFFFF; WRAM(A,r); break;}
#define ST_ZPY(r)	{A=(uint8)(_Y+RMem(_PC)),++_PC,_PC&=0xFFFF; WRAM(A,r); break;}
#define ST_AB(r)	{GetAB(A); WMem(A,r); break;}
#define ST_ABX(r)	{GetABIWR(A,RX); WMem(A,r); break;}
#define ST_ABY(r)	{GetABIWR(A,_Y); WMem(A,r); break;}
#define ST_IX(r)	{GetIX(A); WMem(A,r); break;}
#define ST_IY(r)	{GetIYWR(A); WMem(A,r); break;}
#define ST_IH(r)	{GetIHWR(A); WMem(A,r); break;}

/* remember that "reset" does not initialize like "power" */
/* at least one cart takes advantage of this fact (press reset thirteen times) */
void X6502_Reset(void)
{
	XFrame.intlow = FCEU_IQRESET;
}

void X6502_Power(void)
{
	memset((uint32*)&XFrame,0,sizeof(XFrame));
	X6502_Reset();
	OptFAddr(BWrite[0x4014],DMA4014);
}

void X6502_Init(void)
{
	memset(ZNTable,0,0x80);
	memset(0x80+ZNTable,N_FLAG,0x80);
	ZNTable[0] = Z_FLAG;
}

#if 0 //def env68k
 #undef RMem
 #undef WMem
 #define RMem		RMemCheck
 #define WMem		WMemCheck
#endif

/* reused - optimization purposes */
#define t		cycles

#if (defined(C80x86) && defined(__GNUC__))
 #define regPC		asm ("edi") /* Gives a nice little speed boost. */
#else
 #define regPC
#endif

#define X2A0x_RUN_FUNC	\
	register const uint8 *ZNT=(const uint8*)ZNTable; \
	register uint32	_PC regPC=RPC; \
	register uint32	A,x; \
 \
	_count+=cycles; \
	while(_count>0) { \
		if ((t=_intlow)!=0) { \
			if (t&FCEU_IQRESET) { \
				_S-=3; \
				_PC=_RMem16(0xFFFC); \
				*(int16*)&_PI|=0x101*I_FLAG; \
				_intlow&=~(FCEU_IQRESET|FCEU_JAMMED); \
			} else if (t&FCEU_IQNMI2) { \
				_intlow&=~FCEU_IQNMI2; \
				_intlow|=FCEU_IQNMI; \
			} else if (!(t&FCEU_JAMMED)) \
				if (t&FCEU_IQNMI) { \
					_ADDCYCI(7); \
					PSH(_PC>>8); PSH(_PC); \
					PSH((_P&~B_FLAG)|M_FLAG); \
					_P|=I_FLAG; \
					_PC=NMIVector(); \
					_intlow&=~FCEU_IQNMI; \
				} else if (!(_PI&I_FLAG)) { \
					/* also 0.5+6.5, but there is no FCEU expansion subcycle sync */ \
					_ADDCYC(7); \
					PSH(_PC>>8); PSH(_PC); \
					PSH((_P&~B_FLAG)|M_FLAG); \
					_P|=I_FLAG; \
					_PC=IRSRVector(); \
				} \
			_intlow&=~FCEU_IQTEMP; \
			if (_count<=0) { \
				_PI=_P, _P&=~B_FLAG; \
				goto end; \
			} \
		} \
		if (DebugStepHook) { \
			XFrame.PC=_PC; \
			DebugStepHook(0); \
			_PC=RPC; \
		} \
		_PI=_P, _P&=~B_FLAG; \
		t=RMem(_PC),x=CycTable[t]; \
	exec:	ADDCYC(x); \
		x=_tcount, _tcount=0; \
		if (MapIRQHook) \
			MapIRQHook(x); \
		FCEU_SoundCPUHook(x); \
		++_PC, _PC&=0xFFFF;

/* if not local, then assembly will pick it up */
/* it takes over ten minutes to compile with optimizations on using SC! */
void FP_FASTAPASS(1) X2A0x_RunFull(int32);
void FP_FASTAPASS(1) X2A0x_RunFast(int32);

#ifdef LOCALCODE
/* to get all gg/cheats to work, treat RAM as Mem... it is slower */
 #define RRAM		RMem
 #define WRAM		WMem

/* the N6502 always reads an operand byte, even if it's unused */
/* this is part of why there are no 1-cycle instructions */
/* something to cut down on code by putting this in the loop once */
#define IMP		(void)RMem(_PC)
/* each stack pull commits a pre-read */
#define PUL()		(RRAM(0x100+_S),++_S,RRAM(0x100+_S))

void FP_FASTAPASS(1) X2A0x_RunFull(register int32 cycles)
{
	X2A0x_RUN_FUNC
		switch(t)
		{
		#include "ops6510.h"
		}
	}
end:	XFrame.PC = _PC;
}

 /* only used by Zero Page modes */
 #undef RRAM
 #undef WRAM
 #define RRAM(A)	RAM[A]
 #define WRAM(A,d)	(RAM[A]=d)

 #undef PUL
 #define PUL()		(++_S,RRAM(0x100+_S))

void FP_FASTAPASS(1) X2A0x_RunFast(register int32 cycles)
{
	X2A0x_RUN_FUNC
		switch(t)
		{
		#include "ops6510.h"
		}
	}
end:	XFrame.PC = _PC;
}
#endif

typedef void FP_FASTAPASS(1) (*X650x_RunCall)(int32 cycles);

X650x_RunCall	X650x_Run = X2A0x_RunFast;

void X650x_DebugLevel(int bits)
{
	static int	level;

	if (bits>0)
		level|= bits;
	else
		level&= bits;
	OptFAddr(X650x_Run,level ? X2A0x_RunFull:X2A0x_RunFast);
}
