/*
 sound.c
  N2A0x PSG/pseuedo-Audio Processing Unit and general sound handlers

 * MacFCEU 0.9d6
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  sound.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "WAVE.h"
#include "fsound.h"
#include "filter.h"
#include "state.h"

#include "ftrig.h"
#include "memory.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

/* ---- constants */

#define EnvSpeedMask		0xF

/* rectangle (logical pulse): 4000-4007h */
#define RWaveLnMin		8u
#define RWaveLnMask		0x7FF
#define RWaveLnHigh		0x700 /* was 0xFF00, internally AND'd again */
#define RSweepEnable		0x80
#define RSweepShift		4
#define REnvelopeLoop		0x20 /* looped play not available prior to N2A03 rev. F */

/* triangle: 4008h */
#define TLoadMask		0x7F
#define TStartMask		0x80

/* number of cycles before triangle clocking starts (2015.10.09) */
#define TriLatency		100

/* noise: 400C-400Fh */
#define NShiftShort		0x80

/* 4010h */
#define DMCFEnable		0x80
#define DMCFLoop		0x40
#define DMCFMask		0xF

/* 4015h reads */
#define DMCSIRQ			0x80
#define FrameSIRQ		0x40
#define RdyClSIRQ		0x20

/* 4017h */
#define FMode5Step		0x80
#define FModeNoIRQ		0x40
#define FModeMask		0xC0

static const uint8	lengthtable[32] =
{ /* updated to fit apu_ref.txt - 0x0D*2 (0x1A) on second row was 0x0e*2...was smoother */
	0x0A, 0xFE, 0x14, 0x02, 0x28, 0x04, 0x50, 0x06,
	0xA0, 0x08, 0x3C, 0x0A, 0x0E, 0x0C, 0x1A, 0x0E,
	0x0C, 0x10, 0x18, 0x12, 0x30, 0x14, 0x60, 0x16,
	0xC0, 0x18, 0x48, 0x1A, 0x10, 0x1C, 0x20, 0x1E
};

/* NTSC-based */
static const uint16	GenNoiseFreqTable[16] =
{
	  4,   8,  16,  32,  64,  96, 128, 160,
	202, 254, 380, 508, 762,1016,2034,4068
};
static const uint16	GenDMCTable[16] =
{
	428, 380, 340, 320, 286, 254, 226, 214,
	190, 160, 142, 128, 106,  84,  72,  54
};

static const uint16	PALNoiseFreqTable[16] =
{
	  4,   7,  14,  30,  60,  88, 118, 148,
	188, 236, 354, 472, 708, 944,1890,3778
};
static const uint16	PALDMCTable[16] =
{ /* mostly 1/1.08 that of the NTSC table */
	398, 354, 316, 298, 266, 236, 210, 198,
	176, 148, 131, 118,  98,  78,  66,  50,
};

static const uint16	fIRQCounts[2] = {29831, 33255};

/* CPU, mode, count */
static const uint16	fDividerCounts[2][8] =
{
	{ /* 14915*2 4-step total */
		7457+1, 7459-1, 7456, 7457+1, /* merging IRQ steps in two */
		7458+7452, 7458, 7456, 7458  /* merging 5-step no-op */
	},
	{ /* 16627*2 4-step total (was 16626*2 in FCE Ultra) */
		8312+2, 8315-1, 8314, 8312, /* merging IRQ steps in one */
		8314+8312, 8314, 8314, 8312 /* merging 5-step no-op */
	}
};
static const uint16	*fDivider = *fDividerCounts;

/* ---- globals */

extern uint32 FEmuSpeedRatio;

/* amplitude tables */
static int32	rectwlookup[32], trilookup[32];
/* 15*9+15*9+127 = 397 is adequate */
static int32	tpcmwlookup[398]; /* triangle/noise/DMC -- increased for overflow, was 203, 400 */

/* virtual wave, prior to filtering */
int32		FCEUWave[13472]; /* 111860.7955 Hz/48/17.3% */
/* post-filter results */
static int16	WaveFinal[13472];

EXPSOUND	GameExpSound; /* zeroed out in FCEU.c */

int32		SndTriActive;
static uint8	TriCount;
static uint8	TriReload, TriLatDec;
static int32	TriFreq, TriStep;

uint8		PSGReg[16];

uint8		IRQFrameMode;		/* $4017 xx000000 */
uint8		RawDALatch;		/* $4011 0xxxxxxx */

uint8		EnabledChannels;	/* 4015h */

typedef struct ENVUNIT/*(4)*/
{
	uint8	Speed;
	uint8	reloaddec;//Mode;
	uint8	DecCountTo1;
	uint8	decvolume;
} ENVUNIT;

ENVUNIT		SndEnvUnits[3];

/* NOAC's duty count are in reverse order compared to NTSC */

uint8		SndRecThresh[2];
static int	RecDutyCount[2];
uint8		RecSweepOn[2];
uint32		RecCurFreq[2];
static uint8	SweepCount[2];

static uint16	nreg;
static int32	sqacc[2];
static int32	tracc;
static int32	rnacc;

uint8		Soundfcnt;
int32		Soundfccnt; /* frame cycle count (decremented) */
//int32		Soundfhinc; /* ...increment */

uint32		soundtsoffs;

/* multi-rate variables */
int32		nesincsize = 131072;
uint32		soundtsinc = 65536;

/* length tables */
int32		soundlencount[4];

/* one of the two tables above loaded by SetSoundVariables */
static uint32 CurrentNoiseFreqTable[16];
uint32		CurrentDMCTable[16];
static uint16	rnvol;

// $4010        -        Frequency
// $4011        -        Actual data outputted
// $4012        -        Address register: $c000 + V*64
// $4013        -        Size register:  Size in bytes = (V+1)*64

int32		DMCacc;
int32		DMCPeriod;
uint8		DMCBitCount =-1; /* was 0/-8 */

uint8		DMCAddressLatch, DMCSizeLatch; /* writes to 4012 and 4013 */
uint8		DMCFormat;	/* Write to $4010 */

uint32		DMCAddress = 0;
int32		DMCCounter = 0; /* was DMCSize */

uint8		DMCShiftF[2]; /* first byte kept zero for BE 68K Asm */
#define DMCOutBuf DMCShiftF[1]

uint8		SIRQStat = 0;

uint8		DMCDMABuf = 0;
uint8		DMCBuffered = 0;
uint8		DMCActive = 0;
static uint8	DMCMask;

static void Dummyfunc(void) {};
#if HAVE_HQ
void (*DoRec1)(void)=Dummyfunc;
void (*DoRec2)(void)=Dummyfunc;
void (*DoTriangle)(void)=Dummyfunc;
void (*DoNoise)(void)=Dummyfunc;
void (*DoDMC)(void)=Dummyfunc;
#else
void (*DoRec)(void)=Dummyfunc;
void (*DoPCM)(void)=Dummyfunc;
#define DoRec1			DoRec
#define DoRec2			DoRec
#define DoTriangle		DoPCM
#define DoNoise			DoPCM
#define DoDMC			DoPCM
#endif

#if HAVE_HQ
uint32	FCEUChannelBC[5];
#else
uint32	FCEUChannelBC[3];
#endif

static uint16	*ntable;

/* ---- macros */

#define EORNoise(n,c)		(n=((n>>c^n>>14)&1)+n+n)

#define LoadDMCPeriod(V)	(DMCPeriod=CurrentDMCTable[(uint8)(V)])

#define PrepDMC() \
	(DMCAddress=0x300+(uint32)DMCAddressLatch<<6,DMCCounter=DMCSizeLatch*16u+1)

/* Instantaneous?  Maybe the new freq value is being calculated all of the time... */
/* Beware: this is now a macro */
#define CheckRFreq(c,s)		(s&8 || !(((c>>(s&7))+c)&0x800))

int32		XFStatL;
#define SFStat			((uint8*)&XFStatL)[3*!BIG_ENDIAN] /* top bits */

#if HAVE_HQ
 #define DoRecx(x) \
	if (x) \
		DoRec2(); \
	else \
		DoRec1();
#else
 #define DoRecx(x) \
	DoRec();
#endif

//static void SQReload(int x,
//	uint8 V)
#define RecReload(x,V) do{ \
	if (EnabledChannels&(x+1)) /*(1<<x)*/ { \
		DoRecx(x); \
		if (!(SFStat&16<<x)) /* written on same cycle as envelop decrement */ \
			soundlencount[x] = lengthtable[V>>3]; /*&31];*/ \
	} \
	RecSweepOn[x] = PSGReg[x*4+1]&RSweepEnable; \
	RecCurFreq[x] = PSGReg[x*4+2]|((uint32)V<<8&RWaveLnHigh); \
	SweepCount[x] = (PSGReg[x*4+1]>>RSweepShift&7)+1; \
	RecDutyCount[x] = 7; \
	SndEnvUnits[x].reloaddec = 1; \
}while(0)

/* ---- prototypes */

DECLFW(WritePSG0);
DECLFW(WritePSG1);
DECLFW(WritePSG2);
DECLFW(WritePSG3);
DECLFW(WritePSG4);
DECLFW(WritePSG5);
DECLFW(WritePSG6);
DECLFW(WritePSG7);
DECLFW(WritePSG); /* 8,9,D */
DECLFW(WritePSGA);
DECLFW(WritePSGB);
DECLFW(WritePSGC);
DECLFW(WritePSGE);
DECLFW(WritePSGF);
DECLFW(WriteDMCCtl);
DECLFW(WriteDMCDA);
DECLFW(WriteDMCAddr);
DECLFW(WriteDMCSize);
DECLFW(SndStatusWrite);
DECLFR(SndStatusRead);
DECLFW(WriteIRQFM);

void SndUpdateTriangle(void);

/* ---- code */

static void SndUpdateRecAB(void)
{
	uint8		t;

	t = PSGReg[0];
	t>>= 6, t+= !(t+=t);
	SndRecThresh[0] = t;
	t = PSGReg[4];
	t>>= 6, t+= !(t+=t);
	SndRecThresh[1] = t;
}

void SndUpdateTriangle(void)
{
	SndTriActive =-(FSettings.soundactivech&4 && TriCount && soundlencount[2]);// && TriFreq>2); //4;
	TriFreq = (PSGReg[10]|((uint32)PSGReg[11]<<8&RWaveLnHigh))+1;
}

static INLINE void SndUpdateRNoise(void)
{
	uint32		v;

	if (!soundlencount[3] || !(FSettings.soundactivech&8))
		v = 0;
	else if (PSGReg[12]&16) //(SndEnvUnits[2].Mode&1)
		v = SndEnvUnits[2].Speed;
	else
		v = SndEnvUnits[2].decvolume;
	rnvol = v;
}

/* rect A */
DECLFW(WritePSG0)
{
(void)A;
	DoRec1();
	PSGReg[0] = V;
//	SndEnvUnits[0].Mode = V>>4&3;
	SndEnvUnits[0].Speed = V&EnvSpeedMask;
	V>>= 6, V+= !(V+=V);
	SndRecThresh[0] = V;
}

DECLFW(WritePSG1)
{
(void)A;
	RecSweepOn[0] = V&RSweepEnable;
	PSGReg[1] = V;
}

DECLFW(WritePSG2)
{
(void)A;
	DoRec1();
	RecCurFreq[0]&= RWaveLnHigh;
	RecCurFreq[0]|= V;
	PSGReg[2] = V;
}

DECLFW(WritePSG3)
{
(void)A;
	RecReload(0,V);
	PSGReg[3] = V;
}

/* rect B */
DECLFW(WritePSG4)
{
(void)A;
	DoRec2();
	PSGReg[4] = V;
//	SndEnvUnits[1].Mode = V>>4&3;
	SndEnvUnits[1].Speed = V&EnvSpeedMask;
	V>>= 6, V+= !(V+=V);
	SndRecThresh[1] = V;
}

DECLFW(WritePSG5)
{
(void)A;
	RecSweepOn[1] = V&RSweepEnable;
	PSGReg[5] = V;
}

DECLFW(WritePSG6)
{
(void)A;
	DoRec2();
	RecCurFreq[1]&= RWaveLnHigh;
	RecCurFreq[1]|= V;
	PSGReg[6] = V;
}

DECLFW(WritePSG7)
{
(void)A;
	RecReload(1,V);
	PSGReg[7] = V;
}

/* 8,9,D */
DECLFW(WritePSG)
{
	PSGReg[A&15] = V;
}

/* triangle */
DECLFW(WritePSGA)
{
(void)A;
	DoTriangle();
	PSGReg[10] = V;
	SndUpdateTriangle();
}

DECLFW(WritePSGB)
{
(void)A;
	DoTriangle();
	if (EnabledChannels&4)
		soundlencount[2] = lengthtable[V>>3];
	TriReload = 1;
	PSGReg[11] = V;
	SndUpdateTriangle();
}

/* noise */

DECLFW(WritePSGC)
{
(void)A;
	DoNoise();
	PSGReg[12] = V;
//	SndEnvUnits[2].Mode = V>>4&3;
	SndEnvUnits[2].Speed = V&EnvSpeedMask;
	SndUpdateRNoise();
}

DECLFW(WritePSGE)
{
(void)A;
	DoNoise();
	PSGReg[14] = V;
}

DECLFW(WritePSGF)
{
(void)A;
	DoNoise();
	if (EnabledChannels&8)
		soundlencount[3] = lengthtable[V>>3];
	SndEnvUnits[2].reloaddec = 1;
	PSGReg[15] = V;
	SndUpdateRNoise();
}

DECLFW(WriteDMCCtl)
{
(void)A;
	DoDMC();
	LoadDMCPeriod(V&DMCFMask);
	DMCFormat = V;
	if (SIRQStat&DMCSIRQ)
		if (!(V&DMCFEnable))
		{
			XFrame.intlow&=~FCEU_IQDMC; //X6502_IRQEnd
			SIRQStat&=~DMCSIRQ;
		} /*else -- this wasn't found
			XFrame.intlow|= FCEU_IQDMC; //X6502_IRQBegin
		*/
}

DECLFW(WriteDMCDA)
{
(void)A;
	DoDMC();
	RawDALatch = V&127;
}

DECLFW(WriteDMCAddr)
{
(void)A;
	DMCAddressLatch = V;
}

DECLFW(WriteDMCSize)
{
(void)A;
	DMCSizeLatch = V;
}

void FASTAPASS(1) FrameSoundStuff(uint8 V);
void FrameSoundUpdate(void);
void X2A0xFrameIRQ(void);

#ifdef env68k
void DMCDMA(void);
#else
static
#endif
INLINE void DMCDMA(void)
{ /* rewritten as a separate function; 2011.02.12 */
//	X6502_DMR(DMCAddress); -- 3 DMR cyc in the correct form
	X6502_DMR(DMCAddress); //this cycle should be discounted if from a $4xxx addr
	X6502_DMR(DMCAddress);
	DMCDMABuf = X6502_DMR(DMCAddress);
	DMCAddress = (uint16)(DMCAddress+1)|0x8000;
	DMCBuffered = 1;

	if (!--DMCCounter)
		if (DMCFormat&DMCFLoop)
			PrepDMC();
		else if (DMCFormat&DMCFEnable) /* SIRQStat was prev. set anyway */
		{
			SIRQStat|= DMCSIRQ;
			XFrame.intlow|= FCEU_IQDMC; /*X6502_IRQBegin*/
		}
}

DECLFW(SndStatusWrite) //4015h
{ /* this should now be 100% accurate */
(void)A;
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
	DoDMC();
#else
	DoRec();
	DoPCM();
#endif
	SIRQStat&=~DMCSIRQ;
	XFrame.intlow&=~FCEU_IQDMC; /*X6502_IRQEnd*/
	EnabledChannels = V;//&=31;

	/* Force length counters to 0. */
	if (!(V&1))
		soundlencount[0] = 0;
	if (!(V&2))
		soundlencount[1] = 0;
	if (!(V&4))
		{soundlencount[2] = 0, SndTriActive = 0;}
	if (!(V&8))
		soundlencount[3] = 0;
	if (!(V&16))
		DMCCounter = 0;
	else if (!DMCCounter)
	{
		PrepDMC();
		if (!DMCBuffered) /* 2011.02.12 */
			DMCDMA();
	}
}

uint8 SndGetLenStat(void);
uint8 SndGetLenStat(void)
{
	uint8		ret;

	ret = soundlencount[3]!=0, ret+= ret;
	ret+= soundlencount[2]!=0, ret+= ret;
	ret+= soundlencount[1]!=0, ret+= ret;
	ret+= soundlencount[0]!=0;
	return ret;
}

DECLFR(SndStatusRead) //4015h
{ /* other than top-precision timing, this is accurate */
	uint8		ret;
(void)A;
	if ((ret=SFStat)&0x40)
		ret&= 15; /* 2A0x emu reflects completion; ret will reflect status on previous cycle */
	else
		ret = SndGetLenStat();
	ret+= (DMCCounter!=0)*16u;
	ret+= SIRQStat;
	if (!fceuindbg)
		SIRQStat&=~FrameSIRQ,
			XFrame.intlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd*/
	return ret;
}

DECLFW(WriteIRQFM) //4017h
{
(void)A;
	if (V&FMode5Step)
		{Soundfcnt = 0; FrameSoundUpdate();}
	XFStatL = 0x80000000; /* disable frame IRQ */
	fDivider = fDividerCounts[PAL&1]+((V&FMode5Step)>>7-2);
	Soundfcnt = 1;
	Soundfccnt = fDivider[1]+((int64lo(timestampbase)+timestamp)&1); /* round odd cycle */
	/* no IRQ frame repeat */
	if ((IRQFrameMode=V&FModeMask)==0) /* neither 5-step nor IRQ disable */
		XFStatL = fIRQCounts[PAL]; /* set IRQ clock */
	else if (V&FModeNoIRQ) /* added no-irq check; 2010.06.02 */
		SIRQStat&=~FrameSIRQ,
			XFrame.intlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd*/
}

//if(SweepCount[P]>0) \
//	--SweepCount[P]; \
//if(SweepCount[P]>0) \
//	break;

#define sweepIx(P) do{int32 mod; \
	if(!(PSGReg[P*4]&REnvelopeLoop) && soundlencount[P]>0) { \
		--soundlencount[P]; \
		if(SFStat&0x40) \
			SFStat|=16<<P; \
	} \
	/* Frequency Sweep Code Here */ \
	/* xxxx 0000 */ \
	/* xxxx = hz.  120/(P+1) */ \
	if(!RecSweepOn[P]) \
		break; /* no sweep */ \
	if(SweepCount[P] && --SweepCount[P]) \
		break; \
	SweepCount[P]=(PSGReg[P*4+1]>>RSweepShift&7)+1; /*+1;*/ \
	if(PSGReg[P*4+1]&8) { \
		if(RecCurFreq[P] && (mod=PSGReg[P*4+1]&7)!=0 /*&& RecSweepOn[P]&0x80*/ ) \
			RecCurFreq[P]-=(P^1)+(RecCurFreq[P]>>mod); \
	} else if(((mod=RecCurFreq[P]>>(PSGReg[P*4+1]&7))+RecCurFreq[P])&0x800) \
		RecSweepOn[P]=0,RecCurFreq[P]=0; \
	else if(RecCurFreq[P] && PSGReg[P*4+1]&7) /*&& RecSweepOn[P]&0x80*/ \
		RecCurFreq[P]+=mod; \
}while(0)

//if(SndEnvUnits[P].DecCountTo1>0) \
//	--SndEnvUnits[P].DecCountTo1; \
//if(SndEnvUnits[P].DecCountTo1) \
//	break; \
//SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
//if(SndEnvUnits[P].decvolume || SndEnvUnits[P].Mode&2) \

#define envelopeDecayIx(P,R) do{ \
	if(SndEnvUnits[P].reloaddec) { \
		SndEnvUnits[P].reloaddec=0,SndEnvUnits[P].decvolume=15; \
		SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
		break; \
	} \
	if(SndEnvUnits[P].DecCountTo1 && --SndEnvUnits[P].DecCountTo1) \
		break; \
	SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
	if((V=SndEnvUnits[P].decvolume)!=0 || PSGReg[R]&REnvelopeLoop) \
		SndEnvUnits[P].decvolume=(V-1)&15; \
}while(0)

void FASTAPASS(1) FrameSoundStuff(uint8 V)
{// Linear counter:  Bit 0-6 of $4008
 // Length counter:  Bit 4-7 of $4003, $4007, $400b, $400f
#if HAVE_HQ
 	DoRec1();
	DoRec2();
	DoNoise();
	DoTriangle();
	DoDMC();
#else
	DoRec();
	DoPCM();
#endif
	if (Soundfccnt==-1)
		SFStat|= SndGetLenStat()|0x40;
	if (!(V&1)) /* Envelope decay, linear counter, length counter, freq sweep */
	{
		sweepIx(0); /* rca */
		sweepIx(1); /* rcb */
		if (!(PSGReg[8]&TStartMask) && soundlencount[2]>0) /* tri */
			--soundlencount[2];
		if (!(PSGReg[12]&REnvelopeLoop) && soundlencount[3]>0) /* rns */
			--soundlencount[3];
	}
	/* Now do envelope decay + linear counter. */
	if (TriReload)
		TriCount = PSGReg[8]&TLoadMask;
	else if (TriCount)
		--TriCount;
	if (!(PSGReg[8]&TStartMask))
		TriReload = 0;
	SndTriActive =-(FSettings.soundactivech&4 && TriCount && soundlencount[2]);// && TriFreq>2);

	envelopeDecayIx(0,0);
	envelopeDecayIx(1,4);
	envelopeDecayIx(2,12);
	SndUpdateRNoise();
}

#undef sweepIx
#undef envelopeDecayIx

void FrameSoundUpdate(void)
{
	if (!(uint8)((Soundfcnt&3)|IRQFrameMode))
	{
		if (XFStatL>0)
			XFrame.intlow|= FCEU_IQFCOUNT;
		SIRQStat|= FrameSIRQ;
		if (Soundfccnt>=0-2u)
		{ /* repeat flag */
			FrameSoundStuff(Soundfcnt);
			Soundfcnt = 4;
			return;
		}
	}
	if (Soundfcnt!=4)
		FrameSoundStuff(Soundfcnt);
	++Soundfcnt, Soundfcnt&= 3; /* 3 becomes zero, longest in 5-step */
	Soundfccnt+= fDivider[Soundfcnt];
}

void FrameSoundIRQ(void)
{
	XFrame.intlow|= FCEU_IQFCOUNT, XFStatL = fIRQCounts[PAL];
	SIRQStat|= FrameSIRQ;
}

/* written into 2A0x.a */
void FASTAPASS(1) FCEU_SoundCPUHook(int cycles)
{
	if (!((XFStatL&=0x80FFFFFF)<0) && (XFStatL-=cycles)<0)
		FrameSoundIRQ();
	if ((Soundfccnt-=cycles)<0)
		FrameSoundUpdate(); /* updates Soundfccnt now */
	/* DMCDMA was prev. here */

	if ((DMCacc-=cycles)>0)
		return;
	do
	{
		if (DMCActive)
		{
			uint8	t = (DMCOutBuf<<2&4)-2;

			if (FSettings.SndRate)
			{ /* Unbelievably ugly hack */
				soundtsoffs+= DMCacc;
				DoDMC();
				soundtsoffs-= DMCacc;
			}
			if (!((t+=RawDALatch)&0x80))
				RawDALatch = t;
		}
		DMCOutBuf>>= 1;
		if (!++DMCBitCount)
		{
			DMCBitCount =-8;
			//if (DMCHaveDMA) DMCHaveSample = 1; else...
			if ((DMCActive=DMCBuffered)!=0)
			{
				DMCBuffered = 0;
				DMCOutBuf = DMCDMABuf;
				if (DMCCounter) //&& !DMCBuffered)
					DMCDMA();
			}
		}
	} while((DMCacc+=DMCPeriod)<=0);
}

#if HAVE_HQ
/* This has the correct phase.  Don't mess with it. */
static INLINE void RDoRec(int x)
{
	int32		V;
	uint32		amp;
	int32		rthresh;
	int32		*D;
	int32		currdc;
	int32		cf;
	int32		rc;

	if (RecCurFreq[x]<RWaveLnMin)
		goto endit;
	cf = RecCurFreq[x], rc = PSGReg[x*4+1];
	if (!CheckRFreq(cf,rc))
		goto endit;
	if (!soundlencount[x])
		goto endit;

	if (PSGReg[x*4]&16) //(SndEnvUnits[x].Mode&1)
		amp = SndEnvUnits[x].Speed;
	else
		amp = SndEnvUnits[x].decvolume;
	amp<<= 26;

	//rthresh = RectDuties[PSGReg[x*4]>>6];
	rthresh = PSGReg[x*4]>>6;
	if (!(rthresh+=rthresh))
		rthresh = 1;

	D = &FCEUWave[FCEUChannelBC[x]];
	V = SOUNDTS-FCEUChannelBC[x];

	currdc = RecDutyCount[x];
	cf = (RecCurFreq[x]+1)*2;
	rc = sqacc[x];

	if (V>0)
		do
		{
			if (currdc<rthresh)
				*D+= amp;
			if (!--rc)
				rc+= cf, currdc = (currdc+1)&7;
		} while(++D,--V);

	RecDutyCount[x] = currdc;
	sqacc[x] = rc;
endit:	FCEUChannelBC[x] = SOUNDTS;
}

static void RDoRec1(void)
{
	RDoRec(0);
}

static void RDoRec2(void)
{
	RDoRec(1);
}
#endif

//int tmpamp=0;
//if(RectDutyCount[0]<rthresh[0])
// tmpamp=amp[0];
//if(RectDutyCount[1]<rthresh[1])
// tmpamp+=amp[1];
//tmpamp=wlookup1[tmpamp];
//tmpamp = wlookup1[ ttable[0][RectDutyCount[0]] + ttable[1][RectDutyCount[1]] ];

#define makettable(x,z) do{int _y,_z; \
	_y=RecCurFreq[x],_z=PSGReg[x*4+1]; \
	if(PSGReg[x*4]&16) /*(SndEnvUnits[x].Mode&1)*/ \
		RecThreshTable[x][0]=SndEnvUnits[x].Speed; \
	else \
		RecThreshTable[x][0]=SndEnvUnits[x].decvolume; \
	if(!(FSettings.soundactivech&(x+1)) || RecCurFreq[x]<RWaveLnMin || !CheckRFreq(_y,_z) \
			|| !soundlencount[x]) \
		{z(RecThreshTable[x][0]=0);} \
	/* if(!inie[x]) RecThreshTable[x][0]=0; // Correct? Buzzing in MM2, others otherwise... */ \
	_z=SndRecThresh[x]; /*rthresh[x]=RectDuties[PSGReg[x*4]>>6];*/ \
	/* _y=1..8; _y<_z ? amp:0 */ \
	RecThreshTable[x][1]=(1-_z)>>8&RecThreshTable[x][0]; \
	RecThreshTable[x][2]=(2-_z)>>8&RecThreshTable[x][0]; \
	RecThreshTable[x][3]=(3-_z)>>8&RecThreshTable[x][0]; \
	RecThreshTable[x][4]=(4-_z)>>8&RecThreshTable[x][0]; \
	RecThreshTable[x][5]=(5-_z)>>8&RecThreshTable[x][0]; \
	RecThreshTable[x][6]=RecThreshTable[x][7]=0; /* max 6 */ \
} while(0)

/* new high-quality function */
static void RDoRec(void)
{
	int32		V, end;
	int32		freq0, freq1;
	int32		RecThreshTable[2][8];
	int32		totalout;

	V = FCEUChannelBC[0];
	if ((end=SOUNDTS)<=V)
		return;
	FCEUChannelBC[0] = end;

	/* unrolled */
	makettable(0,(void));
	makettable(1,(void));
	freq0=RecCurFreq[0]+1, freq0+= freq0;
	freq1=RecCurFreq[1]+1, freq1+= freq1;
	/* correct implementation shows duty cycles always count regardless of valid & active */
	totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]] + RecThreshTable[1][RecDutyCount[1]] ];
	do
	{
		FCEUWave[V>>4]+= totalout;

		--sqacc[1];
		if (--sqacc[0]<=0)
		{
			++RecDutyCount[0];
			RecDutyCount[0]&= 7;
			sqacc[0] = freq0; /* set, not add, in case LQ was used */
			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup */
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]]
				+RecThreshTable[1][RecDutyCount[1]] ];
			continue;
		}
		if (sqacc[1]<=0)
		{
		second:	++RecDutyCount[1];
			sqacc[1] = freq1;
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]]
				+RecThreshTable[1][RecDutyCount[1]&=7] ];
		}
	} while(++V<end);
}

static void RDoRecLQ(void)
{
	int32		V, end;
	int32		freq0, freq1;
	int32		inie[2];
	int32		RecThreshTable[2][8];
	int32		totalout;

	V = FCEUChannelBC[0];
	if ((end=((unsigned)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[0] = end;

	/* unrolled */
	inie[0]=inie[1]=nesincsize;
	makettable(0,inie[0]=);
	makettable(1,inie[1]=);

	if (!(RecThreshTable[0][0]|RecThreshTable[1][0]))
		return;
	totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]] + RecThreshTable[1][RecDutyCount[1]] ];
	freq0=(RecCurFreq[0]+1)<<1+17;
	freq1=(RecCurFreq[1]+1)<<1+17;
	do
	{
		FCEUWave[V>>4]+= totalout; //tmpamp;

		sqacc[1]-= inie[1];
		if ((sqacc[0]-=inie[0])<=0)
		{
			do
				++RecDutyCount[0];
			while((sqacc[0]+=freq0)<=0);
			RecDutyCount[0]&= 7;

			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup; 2009.12.27 09:43 EST */
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]]
				+RecThreshTable[1][RecDutyCount[1]] ];
			continue;
		}
		if (sqacc[1]<=0)
		{
		second:	do
				++RecDutyCount[1];
			while((sqacc[1]+=freq1)<=0);
			RecDutyCount[1]&= 7;

			totalout = rectwlookup[ RecThreshTable[0][RecDutyCount[0]]
				+RecThreshTable[1][RecDutyCount[1]] ];
		}
	} while(++V<end);
}

#if HAVE_HQ
static void RDoTriangle(void)
{
	uint32		V;
	int32		tfreq = TriFreq;
	uint32		trout = trilookup[TriStep&31];
	/* frequency check was missing; 2011.02.10 */
	if (!SndTriActive)
	{ /* Counter is halted, but we still need to output. */
		int32	*start = &FCEUWave[FCEUChannelBC[2]];
		int32	count;

		if ((count=SOUNDTS-FCEUChannelBC[2])!=0)
			do
				*start+++= trout;
			while(--count);
		goto end;
	}

	for(V=FCEUChannelBC[2]; V<SOUNDTS; ++V)
	{
		FCEUWave[V]+= trout;
		if (!--tracc)
		{
			tracc = tfreq;
			trout = trilookup[++TriStep&31];
		}
	}
end:	FCEUChannelBC[2] = SOUNDTS;
}

static void RDoNoise(void)
{
	int32		V, nfreq = CurrentNoiseFreqTable[PSGReg[14]&15];
	int32		outo;
	uint32		namptab[2];
	uint16		nr = nreg;

	namptab[1] = 0;
	namptab[0] = rnvol*FSettings.soundnscale;
	outo = namptab[nr>>14];

	if (PSGReg[14]&NShiftShort)
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWave[V]+= outo;
			if (!--rnacc)
				rnacc = nfreq, outo = namptab[(nr=ntable[nr])>>14];
		}
	else
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWave[V]+= outo;
			if (!--rnacc)
				rnacc = nfreq, outo = namptab[(nr=ntable[nr+32768])>>14];
		}
	nreg = nr;
	FCEUChannelBC[3] = SOUNDTS;
}

static void RDoDMC(void)
{
	uint32		V;

	for(V=FCEUChannelBC[4]; V<SOUNDTS; ++V)
		FCEUWave[V]+= (uint32)RawDALatch<<17;
	FCEUChannelBC[4] = SOUNDTS;
}
#endif

static void RDoTriangleNoiseDMC(void)
{
	const uint16	*nt;
	int32		V, end;
	uint32		trout;
	uint32		nr, nfreq, namptab[2];
	uint32		rnout, dmout;
	int32		totalout;

	V = FCEUChannelBC[2];
	if ((end=SOUNDTS)<=V)
		return;
	FCEUChannelBC[2] = end;

	namptab[1] = 0;
	namptab[0] = rnvol*FSettings.soundnscale;

	trout = trilookup[TriStep&31];
	nr = nreg;
	nt = PSGReg[14]&NShiftShort ? ntable:ntable+32768;
	rnout = namptab[nr>>14];
	nfreq = CurrentNoiseFreqTable[PSGReg[14]&15];
	dmout = RawDALatch&DMCMask;

	totalout = tpcmwlookup[trout+rnout+dmout];
	if (!namptab[0])
	{
		if (SndTriActive)
		{
			do
			{
				FCEUWave[V>>4]+= totalout;
				if (--tracc<=0)
				{
					tracc = TriFreq;
					trout = trilookup[++TriStep&31];
					totalout = tpcmwlookup[trout+dmout];
				}
				if (--rnacc<=0)
					rnacc = nfreq, nr = nt[nr];
			} while(++V<end);
			V = 0;
		} else
			do
			{
				FCEUWave[V>>4]+= totalout;
				if (--rnacc<=0)
					rnacc = nfreq, nr = nt[nr];
			} while(++V<end);
		goto end;
	}
	if (SndTriActive)
	{
		do
		{
			FCEUWave[V>>4]+= totalout;
			--rnacc;
			if (--tracc<=0)
			{
				tracc = TriFreq;
				trout = trilookup[++TriStep&31];
				if (rnacc>0)
					goto samp;
				goto nsamp;
			}
			if (rnacc<=0)
			{
			nsamp:	rnacc = nfreq, nr = nt[nr];
				rnout = namptab[nr>>14];
			samp:	totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		V = 0;
	} else
		do
		{
			FCEUWave[V>>4]+= totalout;
			if (--rnacc<=0)
			{
				rnacc = nfreq, nr = nt[nr];
				rnout = namptab[nr>>14];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
end:	nreg = nr;
}

/* noise channel's always the slowest (optimize as much as possible) */
static void RDoTriangleNoiseDMCLQ(void)
{
	int32		V, end;
	uint32		trout, tfreq, tlen;
	uint32		nlen, namptab[2]; /* namptab[1] is zero */
	uint32		rnout, dmout;
	int32		totalout;

	V = FCEUChannelBC[2];
	if ((end=((uint32)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[2] = end;

	tlen = (nlen=nesincsize)&SndTriActive;
	tfreq = TriFreq<<16+1;

	namptab[1] = 0;
	if ((namptab[0]=rnvol*FSettings.soundnscale)==0)
		nlen = 0; /* Quick hack speedup, set [nlen] to 0 */
	trout = trilookup[TriStep&31];
	rnout = namptab[nreg>>14&1];
	dmout = RawDALatch&DMCMask;
	totalout = tpcmwlookup[trout+rnout+dmout];

	if (tlen && nlen)
	{
		const uint16	*nt = PSGReg[14]&NShiftShort ? ntable:ntable+32768;
		uint32		nr = nreg;
		uint32		nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;

		do
		{
			FCEUWave[V>>4]+= totalout;
			rnacc-= nlen;
			tracc-= tlen;
			if (tracc<=0)
			{
				do
					++TriStep;
				while((tracc+=tfreq)<=0);
				trout = trilookup[TriStep&31];
				if (rnacc>0)
					goto samp;
				goto nsamp;
			}
			if (rnacc<=0)
			{
			nsamp:
				do
					nr = nt[nr];
				while((rnacc+=nfreq)<=0);
				rnout = namptab[nr>>14];
			samp:	totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		nreg = nr;
	} else if (tlen)
	{
		do
		{
			FCEUWave[V>>4]+= totalout;
			if ((tracc-=tlen)<=0)
			{
				do
					++TriStep;
				while((tracc+=tfreq)<=0);
				trout = trilookup[TriStep&31];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		V = 0;
	} else if (nlen)
	{
		const uint16	*nt = PSGReg[14]&NShiftShort ? ntable:ntable+32768;
		uint32		nr = nreg, nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;

		do
		{
			FCEUWave[V>>4]+= totalout;
			rnacc-= nlen;
			if (rnacc<=0)
			{
				do
					nr = nt[nr];
				while((rnacc+=nfreq)<=0);
				rnout = namptab[nr>>14];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		nreg = nr;
	} else
		do
			FCEUWave[V>>4]+= totalout;
		while(++V<end);
}

void SetNESSoundMap(void)
{
	SetWriteHandler(0x4000,0x4000,WritePSG0);
	SetWriteHandler(0x4001,0x4001,WritePSG1);
	SetWriteHandler(0x4002,0x4002,WritePSG2);
	SetWriteHandler(0x4003,0x4003,WritePSG3);
	SetWriteHandler(0x4004,0x4004,WritePSG4);
	SetWriteHandler(0x4005,0x4005,WritePSG5);
	SetWriteHandler(0x4006,0x4006,WritePSG6);
	SetWriteHandler(0x4007,0x4007,WritePSG7);
	SetWriteHandler(0x4008,0x400D,WritePSG);
	SetWriteHandler(0x400A,0x400A,WritePSGA);
	SetWriteHandler(0x400B,0x400B,WritePSGB);
	SetWriteHandler(0x400C,0x400C,WritePSGC);
	SetWriteHandler(0x400E,0x400E,WritePSGE);
	SetWriteHandler(0x400F,0x400F,WritePSGF);
	SetWriteHandler(0x4010,0x4010,WriteDMCCtl);
	SetWriteHandler(0x4011,0x4011,WriteDMCDA);
	SetWriteHandler(0x4012,0x4012,WriteDMCAddr);
	SetWriteHandler(0x4013,0x4013,WriteDMCSize);
	SetReadHandler(0x4015,0x4015,SndStatusRead);
	SetWriteHandler(0x4015,0x4015,SndStatusWrite);
	SetWriteHandler(0x4017,0x4017,WriteIRQFM);
}

static int32	inbuf = 0;

int FlushEmulateSound(void)
{
	uint32		b;
	int32		x, end, left;

	if ((x=timestamp)==0)
		return 0;
	if (!FSettings.SndRate)
	{
		inbuf = left = end = 0;
		goto nosnd;
	}
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
	DoDMC();
	if (FSettings.soundq)
	{
		int32	*s = &FCEUWave[soundtsoffs];

		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSFME; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.HiFill[b])
					GameExpSound.HiFill[b]();
		/* ==0 already checked */
		do
			b = *s, *s++ = (b&0x1FFFF)+tpcmwlookup[b>>17&0x1FF]+rectwlookup[b>>26];
		while(--x);
		end = NeoFilterSound(FCEUWave,WaveFinal,SOUNDTS,&left);
		inbuf = end;

		memmove(FCEUWave,FCEUWave+SOUNDTS-left,left*sizeof(uint32));
		memset(FCEUWave+left,0,sizeof(FCEUWave)-left*sizeof(uint32));

		for(b=0; b<=SCSFME; ++b)
			if (GameExpSound.HiSync[b])
				GameExpSound.HiSync[b](left);

		FCEUChannelBC[0] = FCEUChannelBC[1] = FCEUChannelBC[2] =
			FCEUChannelBC[3] = FCEUChannelBC[4] = left;
	}
	else
#else
	DoRec();
	DoPCM();
#endif
	{
		end = ((uint32)SOUNDTS<<16)/soundtsinc;
		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSFME; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.Fill[b])
					GameExpSound.Fill[b](end&15);
		inbuf = end>>4;
		SexyFilter(FCEUWave,WaveFinal,inbuf);

		//if(FSettings.lowpass)
		// SexyFilter2(WaveFinal,end>>4);

		if (end&15) /* roll ending sample to beginning */
			FCEUWave[0] = FCEUWave[inbuf];
		FCEUWave[inbuf] = 0;
	}
nosnd:
#if HAVE_HQ
	if (FSettings.soundq)
		soundtsoffs = left;
	else
	{
		FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
			FCEUChannelBC[2] = FCEUChannelBC[0] = end&15;
		soundtsoffs = (soundtsinc*(end&15))>>16;
#if HAVE_HQ
	}
#endif
	FCEU_WriteWaveData(WaveFinal,inbuf); /* does nothing if sound recording is off */
	return inbuf;
}

int GetSoundBuffer(int16 **W)
{
	*W = WaveFinal;
	return inbuf;
}

/* FIXME:  Find out what sound registers get reset on reset.  I know $4001/$4005 don't,
due to that whole MegaMan 2 Game Genie thing.
*/

void FCEUSND_Reset(void)
{ /* mode is not initialize when fully powered */
	IRQFrameMode = FEmDendy ? FModeNoIRQ:0; /* enabled for standard case */
	fDivider = fDividerCounts[PAL&1];
	Soundfcnt = 1;
	Soundfccnt = fDivider[1]-10; //Soundfhinc;

	nreg = 1;
	//if (!nesincsize) -- ah, just let it dbz
	//	nesincsize = 131072;
	rnacc = sqacc[1] = sqacc[0] = (2048u<<17)/nesincsize;
	tracc = TriLatency;

	RecSweepOn[0] = RecSweepOn[1] = 0;
	RecCurFreq[0] = RecCurFreq[1] = 0;

	//some of these are cleared only on Power
	DMCAddressLatch = DMCSizeLatch = DMCFormat = DMCOutBuf = 0;
	DMCAddress = DMCCounter = DMCBuffered = DMCActive = 0;
	DMCacc = CurrentDMCTable[0]; //1; 2011.02.13

	SIRQStat = 0;
	XFStatL = 0x80000000u;

	RawDALatch = 0;

	TriCount = 0;
	TriReload = 0;
	TriStep = 17;
	EnabledChannels = 0;

	soundlencount[0] = soundlencount[1] = soundlencount[2] = soundlencount[3] = 0;
	SndTriActive = 0;

	/* added; 2011.02.10 */
	WritePSG3(0x4003,0); /* wrote to the wrong memory with basic WritePSG */
	WritePSG7(0x4007,0);
	PSGReg[0] = PSGReg[4] = PSGReg[12] = 16;
	//SndEnvUnits[0].Mode = SndEnvUnits[1].Mode = SndEnvUnits[2].Mode =
	SndEnvUnits[0].reloaddec = SndEnvUnits[1].reloaddec =
		SndEnvUnits[2].reloaddec = 1;
}

void FCEUSND_Power(void)
{
	SetNESSoundMap();
	dwmemset(FCEUWave,0,sizeof(FCEUWave));
	dwmemset(SndEnvUnits,0,sizeof(SndEnvUnits));
	FCEUSND_Reset();

	/* added (actual may be in Reset, but this gives an option); 2009.12.27 */
	DMCBitCount = 0;
#if HAVE_HQ
	FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
		FCEUChannelBC[2] = FCEUChannelBC[0] = 0;

	soundtsoffs = 0;
	LoadDMCPeriod(DMCFormat&DMCFMask);
}

static void SndSetFuncs(uint8 bits)
{
#if HAVE_HQ
	if (FSettings.soundq) /* >=0 */
	{
		OptFAddr(DoRec1,bits&1 ? RDoRec1:Dummyfunc);
		OptFAddr(DoRec2,bits&2 ? RDoRec2:Dummyfunc);
		OptFAddr(DoTriangle,bits&4 ? RDoTriangle:Dummyfunc);
		OptFAddr(DoNoise,bits&8 ? RDoNoise:Dummyfunc);
		OptFAddr(DoDMC,bits&16 ? RDoDMC:Dummyfunc);
		return;
	} else
#endif
	{
		uint8		bq = soundtsinc==65536;
#if HAVE_HQ
		OptFAddr(DoRec1,bits&1 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoRec2,bits&2 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoTriangle,bits&4 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
		OptFAddr(DoNoise,bits&8 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
		OptFAddr(DoDMC,bits&16 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
#else
		OptFAddr(DoRec,bits&3 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoPCM,bits&28 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
#endif
	}
}

static void LoadLUTs(void)
{
	const uint16	*nc, *dc;
	uint32		*na = CurrentNoiseFreqTable, *da = CurrentDMCTable;
	int		x;

	if (PAL)
		nc = PALNoiseFreqTable, dc = PALDMCTable;
	else
		nc = GenNoiseFreqTable, dc = GenDMCTable;
	x = 16;
	do
	{
		*na++ = *nc++;
		if (RateFixFlags&0x40) /* now optional; 2013.05.15 */
			*da++ = *dc++;
		else
			*da++ = umuldiv(*dc++,FEmuSpeedRatio,500000000)+1>>1;
	} while(--x);
}

static int BuildWLUTs(void)
{
	double		numer;
	int		x;

	LoadLUTs();
	/* speed boost: noise table */
	if (!ntable && (ntable=malloc(32768*2*sizeof(uint16)))==NULL)
	{ /* no memory for sound, for that matter... */
		FSettings.SndRate = 0;
		return 0;
	}
	x = 0;
	do
	{
		ntable[x] = (((x>>8^x>>14)&1)+x+x)&0x7FFF;
		ntable[x+32768] = (((x>>13^x>>14)&1)+x+x)&0x7FFF;
	} while(++x<32768);

	{ /* triangle scale table */
		int32	n = FSettings.soundtscale, v;
	#if HAVE_HQ
		if (FSettings.soundq)
			n<<= 17;
	#endif
		x = v = 0;
		do
			trilookup[x] = v, v+=n;
		while(++x<sizeof(trilookup)/sizeof(trilookup[0])/2);
		do
			trilookup[x] = v-=n;
		while(++x<sizeof(trilookup)/sizeof(trilookup[0]));
	}

	/* rectangle filter */
	x = 1;
	numer = (95.52*64)*FSettings.soundrectamp;
	do
		rectwlookup[x] = numer/(8128.0/x+100);
	while(++x<32);
	/* triangle/noise/DMC filter */
	x = 1, numer = (163.67*64)*FSettings.soundtndamp;
	do
		tpcmwlookup[x] = numer/(24329.0/x+100);
	while(++x<sizeof(tpcmwlookup)/sizeof(tpcmwlookup[0]));

	return 1;
}

static void CalcIncSize(void)
{
	soundtsinc = (unsigned)nesincsize;
	if (FEmuSpeedRatio!=1000000000UL && !(RateFixFlags&0x80))
		soundtsinc = umuldiv(soundtsinc,FEmuSpeedRatio,1000000000UL);
	soundtsinc>>= 1;
}

void FCEU_SetSoundVariables(void)
{
	uint64		ratio;
	uint32		Rate;
	int		x;

	if (!FEmuSpeedRatio)
		FEmuSpeedRatio = 1000000000UL;

	rectwlookup[0] = tpcmwlookup[0] = 0;

	if ((Rate=FSettings.SndRate)!=0)
	{ /* build the wave amplitude LUTs */
		if (!BuildWLUTs())
			goto silence;
	} else
	{
	silence:
		SndSetFuncs(0);
		return;
	}
	if (Rate<32)
		Rate = 32;
#if HAVE_HQ
	MakeFilters(Rate>>4);
#endif
	if (PAL)
	{ // 1662607.125 / 16 = 103912.9453125
	extern uint8 FCEU_soundlogrec;
		if (Rate>(int32)PAL_CPU && !FCEU_soundlogrec)
			FSettings.SndRate = Rate = (int32)PAL_CPU;
		_int64xset(ratio,(1L<<17)*(double)PAL_CPU);
	} else
	{ // 1789772.7272.. / 16 = 111860.795454.. | *131072 = 234,589,090,909
		if (Rate>(int32)NTSC_CPU)
			FSettings.SndRate = Rate = (int32)NTSC_CPU;
		_int64xset(ratio,(1L<<17)*(double)NTSC_CPU);
	}
	fDivider = fDividerCounts[PAL&1]+((IRQFrameMode&FMode5Step)>>5);

	for(x=0; x<=SCSFME; ++x)
		if (GameExpSound.RChange[x])
			GameExpSound.RChange[x]();
	sqacc[0] = sqacc[1] = 0;
#if HAVE_HQ
	FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
		FCEUChannelBC[2] = FCEUChannelBC[0] = 0;

	LoadDMCPeriod(DMCFormat&DMCFMask); /* in case of changing from PAL to NTSC */
#if HAVE_HQ
	if (FSettings.soundq)
		nesincsize = 131072;
	else
#endif
	nesincsize = uint64divu32(ratio,Rate);
	CalcIncSize();
	SndSetFuncs(FSettings.soundactivech);

	rnacc = sqacc[1] = sqacc[0] = (2048u<<17)/nesincsize;
	tracc = TriLatency;

	/* recalculate original rate (absolutely needed in case of rate clip) */
	FSettings.OrgRate = FSettings.SndRate;
	if (RateFixFlags&0x80)
		FSettings.OrgRate = umuldiv(FSettings.SndRate,FEmuSpeedRatio,1000000000UL);
}

int FCEUI_Sound(int Rate)
{
	if ((Rate<<4>>4)!=Rate)
		return 0;
	return FCEUI_Sound2(Rate<0 ? 0:Rate<<4);
}

int32 FCEUI_Sound2(int32 Rate)
{
	if (Rate-1u>=(uint32)NTSC_CPU-1u)
		Rate = (uint32)NTSC_CPU&~(Rate>>31);
	FSettings.OrgRate = Rate;
	/* removed 1.6MHz adjustment code */
	/* emulated rate; may be adjusted */
	FSettings.SndRate = Rate;
	FCEU_SetSoundVariables();
	Rate = FSettings.OrgRate;
	FCEUI_SetDCFilterLoss(FSettings.sounddcfl);
	return Rate;
}

/* added; 2008.09.12 */
void FCEUI_SetSoundActiveChannels(uint16 bits)
{ /* flush before setting */
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
#else
	DoRec();
	DoPCM();
#endif
	FSettings.soundactivech = bits;

	if (FSettings.SndRate)
		SndSetFuncs(bits);
	SndUpdateTriangle();
	SndUpdateRNoise();
	DMCMask =-(bits>>4&1);
}

#if HAVE_HQ
void FCEUI_SetLowPass(int q)
{
	FSettings.lowpass = q;
}

void FCEUI_SetSoundQuality(int qty)
{
	FSettings.soundq = (unsigned)qty>255 ? 255:qty;
	FCEU_SetSoundVariables();
}
#endif

void FCEUI_SetSoundVolume(int32 vol)
{
	FSettings.SndVolF = (int32)((uint32)(FSettings.SoundVolume=vol)
		<<(
	#if HAVE_HQ
		FSettings.soundq ? 15:
	#endif
		15-4))/25;
}

/* added; 2010.03.23 */
void FCEUI_SetRectFilterAmp(int16 amp)
{
	FSettings.soundrectamp = amp;
	FCEU_SetSoundVariables();
}

/* added; 2013.05.04 */
void FCEUI_SetTNDFilterAmp(int16 amp)
{
	FSettings.soundtndamp = amp;
	BuildWLUTs();
	CalcIncSize();
}

/* added; 2010.07.17 */
void FCEUI_SetNoiseScale(uint16 scale)
{ /* internal multiplier (no LUT) */
	FSettings.soundnscale = scale;
	SndUpdateRNoise();
}

/* added; 2013.05.05 */
void FCEUI_SetTriangleScale(uint16 scale)
{
	FSettings.soundtscale = scale;
	BuildWLUTs();
}

void FCEUI_SetDCFilterLoss(uint16 loss)
{
	if (FSettings.OrgRate<4)
		FSettings.OrgRate = 4;
	FSettings.sounddcfl = loss;
	FSettings.SndRateF = (loss<<16)/(FSettings.OrgRate>>2);
}

SFORMAT FCEUSND_STATEINFO[] =
{
	{&Soundfccnt,4|FCEUSTATE_RLSB,"FHCN"},
	{&Soundfcnt,1,"FCNT"},
	{ PSGReg,16,"PSG"},
	{&EnabledChannels,1,"ENCH"},
	{&IRQFrameMode,1,"IQFM"},
	{&nreg,2|FCEUSTATE_RLSB,"NREG"},
	{&TriReload,1,"TRIM"},
	{&TriCount,1,"TRIC"},

	{&SndEnvUnits[0].Speed,1,"E0SP"},
	{&SndEnvUnits[1].Speed,1,"E1SP"},
	{&SndEnvUnits[2].Speed,1,"E2SP"},

//	{&SndEnvUnits[0].Mode,1,"E0MO"},
//	{&SndEnvUnits[1].Mode,1,"E1MO"},
//	{&SndEnvUnits[2].Mode,1,"E2MO"},

	{&SndEnvUnits[0].DecCountTo1,1,"E0D1"},
	{&SndEnvUnits[1].DecCountTo1,1,"E1D1"},
	{&SndEnvUnits[2].DecCountTo1,1,"E2D1"},

	{&SndEnvUnits[0].decvolume,1,"E0DV"},
	{&SndEnvUnits[1].decvolume,1,"E1DV"},
	{&SndEnvUnits[2].decvolume,1,"E2DV"},

	{&soundlencount[0],4|FCEUSTATE_RLSB,"LEN0"},
	{&soundlencount[1],4|FCEUSTATE_RLSB,"LEN1"},
	{&soundlencount[2],4|FCEUSTATE_RLSB,"LEN2"},
	{&soundlencount[3],4|FCEUSTATE_RLSB,"LEN3"},
	{ RecSweepOn,2,"SWEE"},
	{&RecCurFreq[0],4|FCEUSTATE_RLSB,"CRF1"},
	{&RecCurFreq[1],4|FCEUSTATE_RLSB,"CRF2"},
	{ SweepCount,2,"SWCT"},

	{&SIRQStat,1,"SIRQ"},
	{&XFStatL,4|FCEUSTATE_RLSB,"SFSL"},

	{&DMCacc,4|FCEUSTATE_RLSB,"5ACC"},
	{&DMCBitCount,1,"5BIT"},
	{&DMCAddress,4|FCEUSTATE_RLSB,"5ADD"},
	{&DMCCounter,4|FCEUSTATE_RLSB,"5SIZ"},
	{&DMCShiftF,2,"5SHF"},

	{&DMCBuffered,1,"5HVD"}, //truncated from 5HVDM, 5HVSP
	{&DMCActive,1,"5HVS"},

	{&DMCSizeLatch,1,"5SZL"},
	{&DMCAddressLatch,1,"5ADL"},
	{&DMCFormat,1,"5FMT"},
	{&RawDALatch,1,"RWDA"},

//	{&sqacc[0],1,"SQA0"},
//	{&sqacc[1],1,"SQA1"},
	{0}
};

void FCEUSND_SaveState()
{
}

void FCEUSND_LoadState(int v)
{
	if (v<9819) /* prev. downshifted 6 */
		IRQFrameMode<<= 6;
	else if (v<9821)
		Soundfcnt>>= 6;
	Soundfcnt&= 3;
	//PSGReg[1]&= 7, PSGReg[5]&= 7;
	DMCBitCount = (DMCBitCount&7)-8;
	LoadDMCPeriod(DMCFormat&DMCFMask);
	DMCAddress = (uint16)DMCAddress|0x8000;
	RawDALatch&= 127;
	DMCShiftF[0] = 0; //for an Asm optimization
	RecCurFreq[0]&= RWaveLnMask, RecCurFreq[1]&= RWaveLnMask;

	SndUpdateTriangle();
	SndUpdateRecAB();
	SndUpdateRNoise();
}
